// classes.C - Steve Ludtke  2/13/98
// Copyright 1998 Baylor College of Medicine
// This impliments k-mean classification, an iterative method
// which finds good spherical clusters for an n dimensional
// distribution

//N classes
//F This implements reference-free k-means classification based on a set of footprint files
//T LC
//1
//P <footprint file>	Input file
//P <# classes>	Number of classes to generate
//P [list]	Produce a list file
//P [weight=<wt file>]	A file containing weighting for the input componenets
//P [maxmove=<n>] The maximum number of particle moves in an iteration before the result is considered stable (default=0)
//P [maxiter=<n>] The maximum number of iterations to run before stopping (default=100)
//P [minsize=<n>] The minumum number of elements to permit in a valid class (default=0)
//D This program is used to classify particles based on a set of invariants produced
//E by one of the 'footprint' programs. These files may optionally have their
//E dimensionality reduced with MSA.

#include <stdio.h>
#include <math.h>
#include "EMData.h"
#include "List.h"
#include "util.h"

#define MINMEM	8		// minimum members in a class to keep it
#define WRITEAV			// this will write out all of the images in each class

struct DOT { float d; unsigned short i1,i2; };

inline int sgn(float x) { if (x==0) return 0; return x<0?-1:1; }

int dotcomp(const void *a, const void *b) { return sgn(*((float *)b)- *((float *)a)); }

main(int argc, char *argv[]) {
double **f;	// these must be double for large data sets
double **cl;
double min=0,max,r,mean,sig,norm;
char s[121],fsp[121],*wtfile=NULL;
int n,n2,nc,bad;
int i,j,k,ii,c,*set,ff=0;
int maxmove=0,maxiter=100,minsize=0;
FILE *in,*out;
EMData *dat;

appinit(argc,argv);

if (argc<3) appusage(argv[0]);

for (i=3; i<argc; i++) {	
	if (strncmp(argv[i],"list",4)==0) ff=1;
	else if (strncmp(argv[i],"weight=",7)==0) wtfile=&argv[i][7];
	else if (strncmp(argv[i],"maxmove=",8)==0) maxmove=atol(&argv[i][8]);
	else if (strncmp(argv[i],"maxiter=",8)==0) maxiter=atol(&argv[i][8]);
	else if (strncmp(argv[i],"minsize=",8)==0) minsize=atol(&argv[i][8]);
	else { printf("Unknown parameter '%s'\n",argv[i]); exit(1); }
}

in=fopen(argv[1],"rb");
if (!in) { printf("Cannot open footprints\n"); exit(0); }
fread(s,120,1,in);
fread(&n,4,1,in);
fread(&n2,4,1,in);
f=(double **)malloc(n2*sizeof(double *));

float *tmp=(float *)malloc(n*sizeof(float));
for (i=0; i<n2; i++) {
	f[i]=(double *)malloc(n*sizeof(double));
	fread(tmp,n,sizeof(float),in);
	for (j=0; j<n; j++) {
		if (!goodf(&tmp[j])) { 
			tmp[j]=0.0;
			printf("Warning, particle %d bad\n",j);
		}
		f[i][j]=tmp[j];
	}
}
fclose(in);
free(tmp);

/*for (j=0; j<n; j++) {
 printf("%d.\t",j);
 for (i=0; i<n2; i++) 
  printf("%f\t",f[i][j]);
 printf("\n");
}*/

int Ref=LOGbegin(argc,argv);
LOG(Ref,argv[1],LOG_INFILE,NULL);
LOG(Ref,"cls0000.hed",LOG_OUTFILE,NULL);

if (wtfile) {
	in=fopen(wtfile,"rb");

	for (i=0; i<n2; i++) {
		
		mean=sig=0;
		
		for (j=0; j<n; j++) {
			mean+=f[i][j];
			sig+=f[i][j]*f[i][j];
		}
		
		mean/=(float)n;
		sig=sig/(float)n-mean*mean;

		fscanf(in," %d %f",&j,&r);
		if (j!=i) printf("Warning, weights out of order %d!=%d!\n",i,j);
		norm=r/sig;

		for (k=0; k<n; k++) f[i][k]*=r;
	}
	
	fclose(in);
}

nc=atoi(argv[2]);
cl=(double **)malloc(n2*sizeof(double *));


// memory for class data points
for (i=0; i<n2; i++) {
	cl[i]=(double *)malloc(nc*sizeof(double));
}

// assign each class to a point in the data set
// since the sets are usually randomly distributed, we just use the
// first data sets
for (i=0; i<nc; i++) {
	for (j=0; j<n2; j++) cl[j][i]=f[j][i];
}

// this holds the info on which class each image is in
set=(int *)malloc(n*sizeof(int));
for (i=0; i<n; i++) set[i]=-1;

int iter=0;
printf("Begin classification\n");
while (1) {
	// classify each image, put each in the class it's closest to
	iter++;
	bad=0;
	for (i=0; i<n; i++) {
		ii=set[i];
		for (j=0; j<nc; j++) {
			r=0;
			for (k=0; k<n2; k++) r+=SQR(f[k][i]-cl[k][j]);
//			printf("%d\t%d\t%lf\n",i,j,r);
			if (r<min || j==0) { set[i]=j; min=r; }
		}
//		printf("%d -> %d  (%lf)\n",i,set[i],min);
		if (ii!=set[i]) bad++;
	}
	
	if (bad<maxmove || iter>maxiter) break;		// any point that moves between classes means
							// we need to keep iterating

//	for (i=0; i<50; i++) printf("%d\t",set[i]);
							
	printf("%d changed\n",bad);

	// now move the center of each class to the mean position of it's members
	int bcls=0;
	for (i=0; i<nc; i++) {
		for (j=0; j<n2; j++) cl[j][i]=0;
	
		c=0;
		for (j=0; j<n; j++) {
			if (set[j]!=i) continue;
			for (k=0; k<n2; k++) cl[k][i]+=f[k][j];
			c++;
		}

		// if we get a class with no members, we randomly move it
		// this will make stabilization take longer, but should
		// give a better distribution
		if (c<minsize && iter+1<=maxiter) { 
			k=random()%n;
			printf("class %d too small (%d) -> %d\n",i,c,k); 
			for (j=0; j<n2; j++) cl[j][i]=f[j][k];
			bcls++;
			continue; 
		}
		for (j=0; j<n2; j++) cl[j][i]/=c;
	
		printf("class %d has %d members\n",i,c);
	}
	printf("%d changed  %d bad classes (iter %d)\n",bad,bcls,iter);
}


// Ok, the classes are stable, write the individual class files

if (!ff) {
	system("/bin/rm cls????.lst");

	dat=new EMData;
	for (i=0; i<n; i++) {
		if (i%10==0) { printf("out -> %d     \r",i); fflush(stdout); }
//		dat->readImage(s,i);
		sprintf(fsp,"cls%04d.lst",set[i]);
//		dat->writeIMAGIC(fsp,-1);
		dat->writeLST(fsp,-1,s,i);
	}
}

// write the classes list file
if (ff) {
	out=fopen("classes.dat","wb");
	for (i=0; i<nc; i++) {
		fprintf(out,"%d\t",i);
		for (j=0; j<n; j++) {
			if (set[j]==i) fprintf(out,"%01d ",j);
		}
		fprintf(out,"\n");
	}
	fclose(out);
}
LOG(Ref,NULL,LOG_END,NULL);

}
