// classesbymra.C ?/98  Steve Ludtke
// Copyright 1998 Baylor College of Medicine

//N classesbymra
//F This program classifies a set of particles based on a set of references (usually projections).
//T LC
//1
//P <particles>	Raw particle filespec	
//P <refrences>	Reference image filespec
//P [split]	Always use this
//P [frac=<num>/<denom>]	Operate only on a fraction of the raw images
//P [projbatches=<num>]	Split the projections into multiple subsets to minimize total memory requirement
//P [verbose[=<n>]]	More verbose output
//P [mask=<rad>]	Mask the classified particles, also reduces noise for better classification
//P [refmask]	This will use a mask generated from the reference image. Assumes the 3D model has been sensibly masked.
//P [imask=<rad>]	Inside mask, classified particles are masked for classification, but not in the cls files.
//P [maxshift=<rad>]	Maximum translation during image alignment
//P [precen]	Particles are assumed to be accurately precentered 
//P [sigfilt]	Apply a real space filter so only high amplitude data is used for classification
//P [tree1=<treefile>]	First stage high speed classification, see <a href=projtree.html>projtree</a>. NOTE : limited parallel support
//P [tree2=<treefile>]	Second stage high speed classification
//P [ctfc]	Projections are CTF corrected before comparison
//P [applyb2]	This will apply B-factors to the projections before comparison. Typically the projections are already filtered, and this may not be desirable.
//P [matrix]	Generates similarity matrix for entire data set in matrix.dat
//P [sep=<nspl>]	Puts each raw particle in 'nspl' different classes.
//P [logit=<label>]	This specifies a label to use in the particle.log file. This records particle number classification throughout a refinement.
//P [touch=<filename>]	Generate an empty file to flag the run completed successfully
//P [rfp]	This will classify using 'rotational footprints' instead of the images themselves. Much faster and quite accruate.
//P [norot]	This will not do any rotational alignment in classification. Primarily used with projections and the matrix option, but may have other uses.
//P [usefilt]	Experimental. Use the 'filt' version of the input file for classification.
//P [refine]	Refines 2D alignment done with fast algorithm. Slower, but not 'slow'.
//P [slow]	Uses a very slow, exhaustive search algorithm for alignment
//P [updateheader]	Save the orientation/center parameters to the image header
//P [flipphase]	Flip phase in addtion to amplitude only when apply CTF or matching amplitude of projections to raw particles
//P [phase]	Experimental. Uses pcmp(), a weighted mean phase-error to compare images for classification
//P [wtphase]	Experimental. Uses pcmp with a SNR weight.
//P [fscmp]	Experimental. Uses fscmp(), a weighted mean phase-error to compare images for classification
//P [wtfscmp]	Experimental. Uses fscmp with a SNR weight.
//P [dfilt]	This will make the rot. avg structure factor of the particle the same as the projection (highly experimental)
//P [frm2d[=<rad>,<samp>]]	This will classify using Fast Rotational Matching 2D. rad is the maximum radius of particles; samp is the number of angular sampling points at each radius circle  
//U  classesbymra start.hed proj.hed split mask=28 ctfc sep=2 tree1=proj.tree
//D NOTE: While classified as experimental, the phase and fscmp options are very useful
//E in most cases these options will produce substantially better results than the default.
//D This classifies a set of raw particles based on a set of reference projections.
//E The output is a set of cls*.img files, one file for each reference image. If the
//E projections were generated from a CTF corrected model, specify the 'ctfc' option
//E so the projections will be 'decorrected' before classification.
//D The 'tree' options implement 2 part classification for higher speed. Check <a href=projtree.html>projtree</a>
//E details on this.
//D The 'sep' option allows each particle to be placed into several different classes. If the
//E images are very noisy, or if the angular spacing is small, misclassification of some
//E particles will inevitably occur. The sep option makes it very likely that one of the
//E classes each particle ends up in will be the correct one. The classalignall procedure
//E can then be used to obtain sufficient consistency within the set of images actually
//E used in each class average.
//D 'frm2d' better used in combination with maxshift, which indicates the 
//E center of mass uncertainty. 

//#define DEBUG

#include <stdio.h>
#include <math.h>
//#include <unistd.h>
#include "EMData.h"
#include "List.h"
#include "util.h"
#include "XYData.h"

// a little class for maintaining an ordered list
class Olist {
public:
float *dot;
int *n;
EMData **ali;
int N;

Olist(int ntokeep) {
	N=ntokeep;
	dot=(float *)malloc(N*sizeof(float));
	n=(int *)malloc(N*sizeof(int));
	ali=(EMData **)malloc(N*sizeof(EMData *));
	for (int i=0; i<N; i++) {
		dot[i]=-1.0e20;
		n[i]=-1;
		ali[i]=NULL;
	}
}

~Olist() {
	free(dot);
	free(n);
	for (int i=0; i<N; i++) delete ali[i];
	free(ali);
}

void insert(float vdot, int vn, EMData *vali) {
	int i,j;

	for (i=0; i<N; i++) if (dot[i]<vdot) break;

	if (i==N) { if (vali) delete vali; return; }

	if (ali[N-1]) delete ali[N-1];
	for (j=N-1; j>i; j--) {
		dot[j]=dot[j-1];
		n[j]=n[j-1];
		ali[j]=ali[j-1];
	}
	dot[i]=vdot;
	n[i]=vn;
	ali[i]=vali;
}

};

main(int argc, char *argv[]) {
int i,j,k,*rn,x,y,nr,ns,ntk=1,split=0,num=0,denom=1,ver=0,sigfilt=0,rfp=0,norot=0;
int n0,n1,shrink=1,mask=-1,imask=0,precen=0,ctfc=0,maxshift=-1,quiet=0,matrix=0,slow=0;
int refine=0,applyb2=0,usepcmp=0,dfilt=0,updateheader=0,flipphase=0;
//int projbatches=1;
int frm2d=0, ccc=0, size=-1, maxshift_input=-1;			//--ycong
int projbatches=1, first=-1, last=-1, projfirst=-1, projlast=-1;
char* scorefile=NULL;
List *data,*dataref,*datarefnc,*newref=NULL;
EMData *d,*d2,*d2f,*drf=NULL,*df=NULL,*dr,*da,*msk,*msk2;
char s[80],t[160],*f2r = argv[1], *touch=NULL;
float dot,classes=-999.0,norm,norm2;
float nullctf[11] = { 0,0,0,0,0,0,0,0,0,0,0 };
float *ctf,*snr=(float *)malloc(10);
float dot_frm=0.0,rmaxH=-1.0f;					//--ycong
char *tree1=NULL,*tree2=NULL,*logit=NULL;
FILE *out;
Olist *best;

FILE *outs;

appinit(argc,argv);

if (argc<3) appusage(argv[0]);

for (i=3; i<argc; i++) {
	if (strncmp(argv[i],"split",5)==0) split=1;
    else if (strncmp(argv[i],"updateheader",12)==0) updateheader=1;
    else if (strncmp(argv[i],"flipphase",9)==0) flipphase=1;
	else if (strncmp(argv[i],"tree1=",6)==0) tree1=&argv[i][6];
	else if (strncmp(argv[i],"tree2=",6)==0) tree2=&argv[i][6];
	else if (strncmp(argv[i],"logit=",6)==0) logit=&argv[i][6];
	else if (strncmp(argv[i],"mask=",5)==0) mask=atoi(&argv[i][5]);
	else if (strncmp(argv[i],"imask=",6)==0) imask=atoi(&argv[i][6]);
	else if (strncmp("refmask",argv[i],7)==0) mask=-2;
	else if (strncmp("maxshift=",argv[i],9)==0) maxshift=atoi(&argv[i][9]);
	else if (strncmp(argv[i],"num=",4)==0) num=atoi(&argv[i][4]);
	else if (strncmp(argv[i],"denom=",6)==0) denom=atoi(&argv[i][6]);
	else if (strncmp(argv[i],"frac=",5)==0) sscanf(&argv[i][5],"%d/%d",&num,&denom);
	else if (strncmp(argv[i],"first=",6)==0) first=atoi(&argv[i][6]);
	else if (strncmp(argv[i],"last=",5)==0) last=atoi(&argv[i][5]);
	else if (strncmp(argv[i],"projbatches=",12)==0) projbatches=atoi(&argv[i][12]);
	else if (strncmp(argv[i],"projfirst=",10)==0) projfirst=atoi(&argv[i][10]);
	else if (strncmp(argv[i],"projlast=",9)==0) projlast=atoi(&argv[i][9]);
	else if (strncmp(argv[i],"shrink=",7)==0) shrink=atoi(&argv[i][7]);
	else if (strncmp(argv[i],"refine",6)==0) refine=1;
	else if (strncmp(argv[i],"precen",7)==0) precen=1;
	else if (strncmp(argv[i],"sigfilt",7)==0) sigfilt=1;
	else if (strncmp(argv[i],"norot",5)==0) norot=1;
	else if (strncmp(argv[i],"quiet",5)==0) quiet=1;
	else if (strncmp(argv[i],"matrix",6)==0) matrix=1;
	else if (strncmp(argv[i],"dfilt",5)==0) dfilt=1;
	else if (strncmp(argv[i],"phase",5)==0) usepcmp=1;
	else if (strncmp(argv[i],"wtphase",7)==0) usepcmp=2;
	else if (strncmp(argv[i],"fscmp",5)==0) usepcmp=3;
	else if (strncmp(argv[i],"wtfscmp",7)==0) usepcmp=4;
	else if (strncmp(argv[i],"ctfc",4)==0) ctfc=1;
	else if (strncmp(argv[i],"slow",4)==0) slow=1;
	else if (strncmp(argv[i],"rfp",3)==0) rfp=1;
	else if (strncmp("frm2d",argv[i],5)==0) {		//ycong
		frm2d=1;
		if (argv[i][5] == '=') sscanf(argv[i]+6,"%f,%d",&rmaxH,&size);
	}
	else if (strncmp("ccc",argv[i],3)==0) ccc=1;		//ycong
	else if (strncmp(argv[i],"sep=",4)==0) ntk=atoi(&argv[i][4]);
	else if (strncmp(argv[i],"applyb2",7)==0) applyb2=1;
	else if (strncmp(argv[i],"usefilt",7)==0) {
		f2r=(char *)malloc(strlen(argv[1])+6);
		strcpy(f2r,argv[1]);
		sprintf(f2r+strlen(argv[1])-4,".filt%s",argv[1]+strlen(argv[1])-4);
	}
	else if (strncmp(argv[i],"verbose",7)==0) {
		if (strncmp(argv[i],"verbose=",8)==0) ver=atoi(&argv[i][8]);
		else ver=1;
	}
	else if (strncmp(argv[i],"usefs=",6)==0) ;
	else if (strncmp(argv[i],"touch=",6)==0) touch=&argv[i][6];
	else if (strncmp(argv[i],"scorefile=",10)==0) scorefile=&argv[i][10];
	else { printf("Unknown parameter '%s'\n",argv[i]); exit(1); }
}

if(frm2d) maxshift_input=maxshift;		// ycong for frm2d

if(projbatches<=0) {
	fprintf(stderr, "WARNING: projbatches=%d, reset to 1\n", projbatches);
	projbatches=1;
}
else if(projbatches>1 && (tree1 || tree2)) {
	fprintf(stderr,"WARNING: projbatches option cannot be specified together with tree1 or tree2, ignored\n");
	projbatches = 1; 	
}

if(projbatches>1 && (projfirst>=0 || projlast>=0)) {
	fprintf(stderr,"WARNING: projbatches and projfirst/projlast can not be both specified. projbatches is ignored\n");
	projbatches = 1; 	
}

if (tree1 && tree2) { printf("Specify either tree1 or tree2, not both!\n"); exit(1); }
if (num>=denom || num<0) { printf("0 <= num < denom !!!\n"); exit(1); }
//if (classes==-999.0 && !split) { printf("Must specify at least 1 of 'classes' or 'split'.\n"); exit(1); }
//if (imask>1 && !precen) { printf("imask may only be used in conjunction with precen\n"); exit(0); }
if (rfp && matrix) error(ERR_WARN,"matrix not properly normalized with rfp");

if (matrix) unlink("matrix.dat");

// READ Reference Images
if (tree1) {
	FILE *in = fopen(tree1,"rb");
	List *ndr = new List;
	char s[256];
	int cn;
	dataref=new List;

	if (!in) { printf("Cannot open '%s'\n",tree1); exit(1); }

	while (fgets(s,255,in)) {
		if (sscanf(s," %d",&cn)!=1) continue;

		d=new EMData;
		d->readImage(argv[2],cn);
		d->unum=cn;
		dataref->addObject(d);

		if (!num) {
			sprintf(s,"tcls%04d.lst",cn);
			if (access(s,F_OK)) continue;
			unlink(s);
//			sprintf(s,"tcls%04d.img",cn);
//			unlink(s);
		}
	}
	fclose(in);
	nr=dataref->count();
}
else if (tree2) {
	FILE *in = fopen(tree2,"rb");
	List *ndr = new List;
	char s[256];
	int cn,cn2,pp[19];
	dataref=new List;

	if (!in) { printf("Cannot open '%s'\n",tree2); exit(1); }

	if (sscanf(argv[1],"tcls%d",&cn2)!=1) {
		printf("tree2 requires 'tcls*' files for input\n"); exit(1);
	}
	while (fgets(s,255,in)) {
		if (sscanf(s," %d",&cn)!=1 || cn!=cn2) continue;

		cn=sscanf(s," %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d",
			pp,pp+1,pp+2,pp+3,pp+4,pp+5,pp+6,pp+7,pp+8,pp+9,pp+10,pp+11,
			pp+12,pp+13,pp+14,pp+15,pp+16,pp+17,pp+18);

		if (cn<4) { printf("Tree file appears strange, '%s'\n",s); exit(1); }

		for (i=0; i<cn; i++) {
			d=new EMData;
			d->readImage(argv[2],pp[i]);
			d->unum=pp[i];
			dataref->addObject(d);
		}
		break;
	}
	fclose(in);
	nr=dataref->count();
}
else {
	nr=fileCount(argv[2]);
}

ns=fileCount(f2r);
printf("%d> %d images to classify.\n",num,ns);

// Read the file containing the list of images to exclude
off_t zfl;
char *zfile=readTextFile("start.exclude",&zfl);
char *zapped=NULL;
if (zfile) {
	if ((strcmp(argv[1],"start.hed")==0 || strcmp(argv[1],"start.img")==0)) {
		zapped=(char *)calloc(1,ns);
		printf("Using image exclusion file\n");
		char *zloop=zfile;
		while (1) {
			int ii,jj;
			if (!sscanf(zloop," %d%n",&ii,&jj)) break;
			zloop+=jj;
			zapped[ii]=1;
		}
	}
	free(zfile);
}

if (matrix) {
	if (!num) {
		int tt[3];
		tt[0]=0;
		tt[1]=nr;
		tt[2]=ns;
		FSsend(FS_BIN,12,"matrix.dat",(char *)tt);
	}
	else sleep(1); // stupid hack to avoid possible file contention
}

rn=(int *)malloc(nr*sizeof(int));
if(tree1||tree2) for (i=0; i<nr; i++) rn[i]=((EMData *)dataref->objectAt(i))->unum;
else for (i=0; i<nr; i++) rn[i]=i;	// rn stores the projection id
	
int Ref=0;
if (!quiet) {
	Ref=LOGbegin(argc,argv);
	LOG(Ref,argv[2],LOG_INFILE,NULL);
	LOG(Ref,argv[1],LOG_INFILE,NULL);
}

d=new EMData;

d->readImage(argv[2],0);
d->edgeNormalize(1);
d->applyMask(mask,4);
d->meanShrink(shrink);
		
x=d->xSize();
y=d->ySize();

msk=d->copy();
msk->setParent(NULL);
msk->one();
msk->applyMask(mask/shrink,4);
msk->applyMask(imask/shrink,5);

msk2=d->copy();
msk2->setParent(NULL);
msk2->one();
msk2->applyMask(mask/shrink,4);
msk2->applyMask(imask/shrink,5);

delete d;

norm=x*y/1000.0;
norm2=PI/(4.0*x*y);
if (maxshift<0) {
	if (mask<=0) maxshift=y/8; else maxshift=y/2-mask/shrink;
}

n0=num*ns/denom;
n1=(num+1)*ns/denom;
if(first>=0) n0 = first;
if(last>n0 && last<=ns) n1=last;

printf("%d> processing %d - %d of %d\n",num,n0,n1,ns);

// Thses are used to store the values while the program runs
// The results are then written out all at once at the end of the
// job. This is done to make the job restartable and to concentrate
// network requests
int *rns = (int *)malloc((n1-n0)*ntk*sizeof(int));
float *dots = (float *)malloc((n1-n0)*ntk*sizeof(float));
float *ort = (float *)malloc((n1-n0)*ntk*sizeof(float)*4);

fflush(stdout);
if (mask==-1) mask=-shrink;
if (!quiet) LOG(Ref,"cls0000.lst",LOG_OUTFILE,NULL);

float *c1,*c2;
if (dfilt) {
	c1=(float *)malloc(sizeof(float)*4*x);
	c2=(float *)malloc(sizeof(float)*4*x);
}

float *sim = (float *)malloc(sizeof(float)*nr);
Olist **bests = (Olist**)malloc((n1-n0)*sizeof(Olist*));

for(int projbatch=0; projbatch<projbatches; projbatch++){
	
int cnr;
int projn0 = projbatch * nr / projbatches;
int projn1 = (projbatch + 1) * nr / projbatches;
if(projfirst>=0) projn0 = projfirst;
if(projlast>projfirst) projn1 = projlast;
if (projn1 > nr) projn1 = nr ;

if(! (tree1 || tree2) ){
	printf("%d> Read %d (%d-%d of %d) references\n", num,projn1-projn0, projn0, projn1-1,  nr );

	dataref=readImages(argv[2],projn0,projn1-1);		// read input projection subsets, inclusive for start/end indexes

	if (dataref==NULL) error("Cannot read input",ERR_EXIT);

	cnr = dataref->count();

	for (i=0; i<cnr; i++) {
		d=(EMData *)dataref->objectAt(i);
		d->unum=i+projn0;
	}
}
else {
	cnr = dataref->count();	
}

for (i=0; i<cnr; i++) {
	d=(EMData *)dataref->objectAt(i);
	j=d->unum;
	if (sigfilt) d->realFilter(3,d->Sigma(),-d->Sigma());
	d->edgeNormalize(1);
//	d->normalize();
	d->applyMask(mask,4);
	d->meanShrink(shrink);
	d=d->copy();
	d->unum=j;
	dataref->replaceObjectAt(i,d);
	(d->makeRFP(1))->normalize();
}

datarefnc=dataref;
if (ctfc || dfilt) {
	dataref=new List;
	for (i=0; i<cnr; i++) {
		delete ((EMData *)datarefnc->objectAt(i))->Parent();
		((EMData *)datarefnc->objectAt(i))->setParent(NULL);
		dataref->addObject(((EMData *)datarefnc->objectAt(i))->copy());
		((EMData *)dataref->objectAt(i))->setCTF(nullctf);
	}
}

printf("Preprocessing complete\n");

/**************************************************
          parameter settings for frm2d 
		9/27/06  ycong  
***************************************************/
float p_max, rmaxL;
float com_with_x=0.0, com_with_y=0.0, com_self_x=0.0, com_self_y=0.0;
float r_size;		// possible r value, to allocate h_haH array
float MAXR;		// maximum r values for sampling the densities
float tstep=1.0;	// radius (linear) search step size (in pixel), default is 1 pix
float angstep;		// angular sampling points at each radius circle, i.e. 32, 64, 128, 256
float rhomax=-0.1, rhomin=0.0;			// not really in use in the current version
float *cen_mass;	//for storage the center of mass of reference images
//float *sum_samp;
float *Int_g;
//float *com_pre;
float *raw_hhat = 0;
//FILE *in_com;	

if(frm2d){
	if (size<0 || shrink) {
		int do360=1;
		size = (int) floor(do360*PI*y/4);	//considering shrink, use y instead of d->ySize()
		size = bestfftsize(size);
	}
	
	if(maxshift_input > 0) {
		p_max=ceil((float)maxshift_input/(float)shrink);
		if(p_max > maxshift) p_max=maxshift;
	}
	else p_max=maxshift; 
	
	if(rmaxH < 0){
		if(mask<=0) rmaxH=y/2.0;
		else rmaxH=float(mask/shrink);
	}
	else if(ceil(rmaxH) > mask && mask>0) {
		fprintf(stderr,"WARNING: the maximum radius of particles (%6.1f) should be <= the radius of mask (%d), reset to the mask radius \n",rmaxH,mask);	
		rmaxH = float(mask/shrink);
	}
	else if(float(rmaxH/shrink) > y/2.0) {
		fprintf(stderr,"WARNING: the maximum radius of particles (%6.1f) should be <= half of the box size (%d), reset to half of the box size \n",float(rmaxH/shrink),y/2);
		rmaxH = y/2.0;
	}
	else rmaxH = float(rmaxH/shrink);
		
	rmaxL = rmaxH;			
	r_size=rmaxH+p_max;			 
	
	if(mask<=0) MAXR=y/2.0+p_max;
	else        MAXR=float(mask/shrink)+p_max;
	
	angstep=size;
	//printf("size=%d, p_max/maxshift=%f,%d, rmaxH=%f, r_size=%f, MAXR=%f, y=%d \n",size,p_max,maxshift,rmaxH,r_size,MAXR,y);
	
	//float rhomax=-0.1, rhomin=0.0;		// not really in use in the current version
	/*if(rmaxH >= rmaxL)
		rhomax=rmaxH-rminL;			// rhomax -> maximum value of rho 
	else
		rhomax=rmaxL-rminH;
	if(p_max < rhomax) rhomax = p_max;
	*/
	
	cen_mass=new float[(nr+1)*2];		
	if(cen_mass == NULL){
		cout <<"can't allocate more memory, terminating 001. \n";
		exit(1);
	}
	
	float *sum_samp = new float[((int)MAXR+1)*size];
	if(sum_samp == NULL ){
		cout <<"can't allocate more memory, terminating 002. \n";
		exit(1);
	}
	for (int m=0;m<size*(int)(MAXR+1);m++)
		sum_samp[m] = 0.0;
	
	int max_ext=0;
	for (int j=0; j<nr; j++) {
		EMData *d_ref = (EMData *)dataref->objectAt(j);
		//d_ref->norm_frm2d();		//reference image has been norm & shrinked in previous eman preprocess 10/28/06
		if(j==0) max_ext=(floor)(sqrt((float)(d_ref->xSize()*d_ref->xSize()+d_ref->ySize()*d_ref->ySize()))+0.5);
		
		float com_x=0.0f, com_y=0.0f;
		d_ref->COM(1, com_x, com_y); 
		//printf("com_x,com_y=%f,%f\n",com_x,com_y);
	
		EMData* ref_pcs = d_ref->unwrap_largerR(0,(int)MAXR,size,rmaxH);
	
		//deletion of d_ref will be taken care of by objectAt
	
		float *data_ref=ref_pcs->getDataRO();
		for (int r=0;r<=(int)rmaxH;r++)
			for (int k=0;k<size;k++) {
				sum_samp[k+r*size] += data_ref[k+r*size];
				//printf("j=%d, r=%d, k=%d, %f, %f  \n ",j,r,k, sum_samp[k+r*size], data_ref[k+r*size]);
			}
		ref_pcs->doneData();
		if (ref_pcs) { 
			delete ref_pcs; 
			ref_pcs=0;
		}
		data_ref = NULL;
	}
	
	Int_g = new float[max_ext+1];
	if(Int_g == NULL ){
		cout <<"can't allocate more memory, terminating 003. \n";
		exit(1);
	}
	outs=fopen("Ref_r_ave_FRM.txt","w"); 
	for(int r=0;r<=max_ext;r++) {
		Int_g[r]=0.0f;
		if (r<=(int)MAXR){
			for (int i=0;i<size;i++) {
				Int_g[r] += sum_samp[i+r*size]/nr;		
			}
		}
		fprintf(outs, "%f\n",Int_g[r]);
	}
	fclose(outs);
	delete[] sum_samp; 
		
	/*com_pre = new float[(ns+1)*4];
	if(com_pre == NULL ){
		cout <<"can't allocate more memory, terminating 004. \n";
		exit(1);
	}
	//FILE *in_com;	
	in_com=fopen("com_predeterm.txt","r");
	if(in_com) { 
		for(int n=0;n<ns;n++) {
			int num=0, yes=0;
			float xx=0.0,yy=0.0;
	
			fscanf(in_com, "%d, %f, %f, %d \n",&num, &xx,&yy, &yes); 
			com_pre[0+n*4]=(float)(num);			//raw image number
			com_pre[1+n*4]=(float)(floor(xx + 0.5));	//com x
			com_pre[2+n*4]=(float)(floor(yy + 0.5));	//com y
			com_pre[3+n*4]=(float)(yes);			//yes=0, use predet. com; yes=1, determine com here using the code 
			//printf("n=%d, %f, %f, %f, %f \n ",n, com_pre[0+n*4],com_pre[1+n*4], com_pre[2+n*4],com_pre[3+n*4]);
		}
		fclose(in_com);
	} */
	//float* raw_hhat = 0;
}
//~~~~~~~~~~~~ ycong ~~~~~~~~~~~~

d = new EMData;
// align and classify
for (i=n0; i<n1; i++) {
	if (zapped && zapped[i]) continue;	// This particle has been marked as bad
	if(projbatch==0) {
		best=new Olist(ntk);
		bests[i-n0]=best;
	}
	else {
		best=bests[i-n0];
	}

	d->readImage(f2r,i);
// we set the B-factor to zero so ctf correction doesn't end up applying it twice
// This won't work properly for CTFF images obviously. still needs to be dealt with...
	if (d->hasCTF() && !applyb2) {
		float *par = d->getCTF();
		par[1]=0;
	}
	if ( (!d->hasCTF()) && (usepcmp==2 || usepcmp==4)) {
		fprintf(stderr, "ERROR: option \"wtphase\" and \"wtfscmp\" only work on images with CTF parameters.");
		exit(1);
	}
	d->edgeNormalize(1);
//	d->normalize();
	d->meanShrink(shrink);
	d->mult(msk2);
	if (d->xSize()!=x ||d->ySize()!=y) { printf("Bad image size %d,%d != %d,%d\n",d->xSize(),d->ySize(),x,y); exit(1); }

	d2=d->copy();
	d2->makeRFP(1);
	EMData *f=d2->doFFT();
	if (dfilt) f->calcRadDist(x,0,.5,c1);

//	d2f=d->copy(0,0);
//	d2f->hFlip();
/*	if (ver>=9) {
		d2->writeIMAGIC("align.hed",-1);
		df->writeIMAGIC("align.hed",-1);
	}*/

//---------------ycong---------------
	if(raw_hhat != 0) { 
		free(raw_hhat); 
		raw_hhat = 0;
	}
	if (frm2d){
		EMData *d2_frm=d2->copy();			//for raw particle image
		//d2->writeIMAGIC("d2.hed",-1);
		/*if(in_com && com_pre[3+i*4] == 0.0) {
			
			float dxx = com_pre[1+i*4];
			float dyy = com_pre[2+i*4];
			com_with_x = d2_frm->xSize()/2 - dxx;
			com_with_y = d2_frm->ySize()/2 - dyy;
			
			d2_frm->setTAlign(dxx,dyy,0);
			d2_frm->fastTranslate();			
		}		
		else { */		
			d2_frm->COM(3, com_with_x, com_with_y);		// intonly=3: no translation of image in the COM function
			d2_frm->exact_COM(1,Int_g, com_with_x, com_with_y,rmaxH);	// purpose: to match the style of function "exact_com"
		//}
		//printf("COM_raw2 %f, %f\n", com_with_x, com_with_y);
		
		EMData *withpcs = d2_frm->unwrap_largerR(0,MAXR,size,rmaxH);
		delete d2_frm;
		EMData *withpcsfft = withpcs->oneDfftPolar(size, rmaxH, MAXR);
		delete withpcs;
		
		raw_hhat = withpcsfft->makeFRM2D_H_hat(rhomax, rhomin, rmaxH, size, tstep, r_size);
		delete withpcsfft;
	}	
//~~~~~~~~~~~~ ycong ~~~~~~~~~~~~

	for (j=0; j<cnr; j++) {
		dr=(EMData *)dataref->objectAt(j);
		if (dfilt || ((d->hasCTF()||d->hasCTFF()) && ctfc && d->CTFCmp(dr))) {
			EMData *dt;
			// if several images in a row have the same ctf, don't recalculate
			if (d->hasCTF()) ctf=d->ctfCurve(flipphase?0:2);
			dt=((EMData *)datarefnc->objectAt(j))->doFFT()->copy();
			if (dfilt) { // match filter and ctf sign
				dt->calcRadDist(x,0,.5,c2);
				for (int jj=0; jj<x; jj++) {
					if (c1[jj]>0 && c2[jj]>0) c2[jj]=sqrt(c1[jj]/c2[jj]);
					else c2[jj]=0;
					//if(ctf[jj]<0)c2[jj]*=-1.;
				}
				if (d->hasCTF()) {
					for (int jj=0; jj<x; jj++) if(ctf[jj]<0)c2[jj]*=-1.;
				}
				dt->applyRadFn(x,0,.5,c2);
			}
			else if (d->hasCTF())  { // ctf sign
				dt->applyRadFn(CTFOS*d->ySize()/2,0,1.0/CTFOS,ctf,0);
			}
			dr=dt->doIFT();
			dt->gimmeFFT();
			dr->edgeNormalize();
//			dr->normalize();
			dr->mult(msk);
			dr->setParent(NULL);
			dr->setCTF(d->getCTF());
			if (mask==-2) {
				EMData *cp=((EMData *)datarefnc->objectAt(j))->copy(0,0);
				cp->realFilter(6);	//value!=0 -> 1
				dr->mult(cp);
				delete cp;
			}
			(dr->makeRFP(1))->normalize();
			delete ((EMData *)dataref->replaceObjectAt(j,dr));
			delete dt;
			if (d->hasCTF()) free(ctf);
			if (usepcmp==2 || usepcmp==4) {
				int nctf=(int)ceil(CTFOS*sqrt(2.0)*d->ySize()/2)+2;
				free(snr);
				snr=d->ctfCurve(3);
			}
		}
		//if (drf) delete drf;
		drf=dr->copy(0,0);
		drf->hFlip();
		if (rfp) {
			df=d2->copy(0,0);
			df->hFlip();
			df->makeRFP(1);
			dot=d2->rotAlign(dr);
			float dot2=df->rotAlign(dr);
			if (dot2>dot) dot=dot2;
			da=NULL;
			dot/=10;
			delete df;
		}
		else {
			if (precen) {
				Euler e = *dr->getEuler();
				da=dr->RFAlign(d2,NULL,imask/shrink);
				dr->setRAlign(&e);
			}
			//----ycong----
			else if (frm2d) {
				Euler e = *dr->getEuler();
				//if(!dfilt) {				//shouldn't do edgeNormalize again, have been
					//dr->edgeNormalize();		//edgenormalized and shrinked in preprocessing
				//}
				EMData *dr_frm0=dr->copy();
				//dr_frm0->writeIMAGIC("refimag.hed",-1);
				
				if(i==n0) {
					dr_frm0->COM(1, com_self_x, com_self_y);	// in COM function, the object was trans. to the COM

					cen_mass[j*2+0]=com_self_x;
					cen_mass[j*2+1]=com_self_y;
				}
				else{
					com_self_x=cen_mass[j*2+0];
					com_self_y=cen_mass[j*2+1];
				}
				delete dr_frm0;
				//printf("# %d  COM_ref %f, %f\n",j, com_self_x,com_self_y);
				
				int neg_rt=1;
				//use fcscmp() to check the similarity of two images  6/26/04
				da=dr->frm2dAlignFlip(drf, dot_frm, d2,0, raw_hhat, rmaxL, MAXR, tstep, angstep, r_size, com_self_x, com_self_y, com_with_x, com_with_y,neg_rt);
				if(refine) da->refineAlign(d2);
				//printf("No.=%d, Corr=%g, self.ang=%f, %f, %f \n\n",j,dot_frm,da->alt()*180.0/3.1415926, da->Dx(), da->Dy() );
						
				dr->setRAlign(&e);
			}
			//~~~~~~ ycong ~~~~~~

			else if (norot) { 
				da=dr->copy(); 
				da->transAlign(d2,1,1,maxshift); 
				da->fastTranslate(); 
			}
			else {
				Euler e = *dr->getEuler();
				if (slow) da=dr->RTFAlignBest(d2,drf,maxshift);
				else da=dr->RTFAlign(d2,drf,1,maxshift);
				if (refine) da->refineAlign(d2);
				//printf("No.=%d, Corr=%g, self.ang=%f, %f, %f \n\n",j,dot_frm,da->alt()*180.0/3.1415926, da->Dx(), da->Dy() );
				dr->setRAlign(&e);
			}
			if (!da) { delete drf; continue; }
			da->mult(msk);
//			dot=da->dot(dr);
#ifndef ALTCMP
//			dot=da->lcmp(d2,0);	// this seems to work VERY poorly, not sure why?
			if (usepcmp==1) dot=da->pcmp(d2,NULL);
			else if (usepcmp==2) dot=da->pcmp(d2,snr);
			else if (usepcmp==3) dot=1.0-da->fscmp(d2,NULL);
			else if (usepcmp==4) dot=1.0-da->fscmp(d2,snr);
			else dot=da->lcmp(d2,1);
#else
			dot=(da->ySize()/2-da->fscmp(d2))/(float)da->ySize();
#endif

//if (j==15) dr->makeRFP()->writeIMAGIC("align.hed",-1);

			// This is a new normalization, basically equivalent to lcmp. The problem this corrects is
			// That 'edgenormalize' makes direct comparison of dot products problematical
//			dot=(da->Sigma()+SQR(da->Mean()))*da->xSize()*da->ySize()+(dr->Sigma()+SQR(dr->Mean()))*dr->xSize()*dr->ySize()-2*dot;
//			dot/=da->xSize()*da->ySize();
		}
		delete drf;
		
#ifdef DEBUG2
		EMData *DDB,*DDR;
		DDB=d2->copy(0,0);
		DDR=da->copy(0,0);
		DDB->writeIMAGIC("mradbug.hed",-1);
		DDR->writeIMAGIC("mradbug.hed",-1);
		DDR->normalizeTo(DDB,1);
		DDB->subtract(DDR);
		if (mask==-2) { DDR->realFilter(6); DDB->mult(DDR); }
		DDB->writeIMAGIC("mradbug.hed",-1);
//		printf("%d. %f\n",j,DDB->Sigma());
		delete DDB;
		delete DDR;
#endif
//		da->subtract(dr);
//		da->writeIMAGIC("t.hed",-1);

		if (da) sim[j+projn0]=dot;
		else sim[j+projn0]=0;

//		if (sim[j]<0) sim[j]=0;
		if (ver>1) printf("%d> %d\t%f\t%f\t %1.1f,%1.1f  %7.3f\t%1.5f  %1.5f  %c\n",num,
			j,dot,sim[j+projn0],da->Dx(),da->Dy(),da->alt()*180./PI,da->Sigma(),dr->Sigma(),da->isFlipped()?'F':' ');

		if (ver>=9) {
//			d2->makeRFP()->writeIMAGIC("align.hed",-1);
//			df->makeRFP()->writeIMAGIC("align.hed",-1);
			d2->writeIMAGIC("align.hed",-1);
//			df->writeIMAGIC("align.hed",-1);
			dr->writeIMAGIC("align.hed",-1);
			da->writeIMAGIC("align.hed",-1);
		}

		if (!rfp) dot=(2.0-dot)*500.0;	// this is for historical purposes
		if (ccc)  dot=dot_frm;	       	// frm2d, use fcs similarity cri.  --ycong
		best->insert(dot,j+projn0,da);	// note that this may delete da
		da=NULL;

//		if (ver) { putchar('0'+(int)floor(dot/norm/50)); fflush(stdout); }
	}

	free(raw_hhat);
	raw_hhat=NULL;		//~~~ycong Jul.31,04

	if (matrix && projbatch == projbatches-1) {
		int *tt = (int *)malloc((2+nr)*sizeof(int));
		tt[0]=-1;
		tt[1]=i;
		memcpy(tt+2,sim,nr*sizeof(int));
		FSsend(FS_BIN,(2+nr)*sizeof(int),"matrix.dat",(char *)tt);
		free(tt);
	}

//	if (ver) putchar('\n');

/*	if (best->dot[0]<0) {
		printf("No class for %d\n",i);
		delete d2;
		delete df;
//		delete d2f;
		continue;
	}*/

	delete d2;
//	delete d2f;
	if (df) { delete df; df=NULL; }

	// class average for dots better than a threshold
//	if (classes!=-999.0) {
//		if (bestdot/norm>classes) {
//			d2=(EMData *)newref->objectAt(best);
//			d2->add(dbest);
//		}
//	}

	// split into class files
	if ( (split || updateheader || scorefile) && projbatch == projbatches-1) {
/*		if (shrink>1) d->readImage(argv[1],i);
		printf("%d -> ",i);
		for (j=0; j<ntk; j++) {
			printf("\t%d (%1.2f)",rn[best->n[j]],best->dot[j]);
			sprintf(s,"%scls%04d.hed",tree1?"t":"",rn[best->n[j]]);
			d->writeIMAGIC(s,-1);
		}
		printf("\n");
		fflush(stdout);*/

		printf("%d> %d -> ",num,i);
		for (j=0; j<ntk; j++) {
			printf("\t%d (%1.2f)",rn[best->n[j]],best->dot[j]);
			//rns[(i-n0)*ntk+j]=rn[best->n[j]];
			rns[(i-n0)*ntk+j]=best->n[j];	// this is the index for refs in memory, should use rn[best->n[j]] to get true index for projection images on disk
			dots[(i-n0)*ntk+j]=best->dot[j];
			float q0=-best->ali[j]->Dx();
			float q1=-best->ali[j]->Dy();
			float qa=-best->ali[j]->alt();
			ort[((i-n0)*ntk+j)*4  ]=qa;
			ort[((i-n0)*ntk+j)*4+1]= q0*cos(qa)+q1*sin(qa);
			ort[((i-n0)*ntk+j)*4+2]=-q0*sin(qa)+q1*cos(qa);
			ort[((i-n0)*ntk+j)*4+3]=best->ali[j]->isFlipped();
		}
		printf("\n");
		fflush(stdout);

	}
	else printf("%d> %d.\n",num,i);

	if(projbatch==projbatches-1) delete best;
}

// save the resulted orientation/center parameters to the image header
if (updateheader && projbatch == projbatches-1) {
	char* imgid = file_basename(argv[1]);
	printf("updating the images header of %s\n", imgid);
	for (i=n0; i<n1; i++) {
		if (zapped && zapped[i]) continue;
		if (d->readImage(argv[1], i, 1)) {	// read in the original image header only to avoid messing up CTF due to shrink
			fprintf(stderr,"%s: error in reading image %s:%d\n",argv[0],argv[1], i);
		}

		int refid = rns[(i-n0)*ntk];	// only take the best score, the reference id in dataref list (can be different from projections index due to tree options)
		int proj_id = rn[refid];	// only take the best score
		float alt = ((EMData *)dataref->objectAt(refid-projn0))->alt();
		float az  = ((EMData *)dataref->objectAt(refid-projn0))->az();
		float phi = ((EMData *)dataref->objectAt(refid-projn0))->phi() + ort[((i-n0)*ntk)*4];
		if (ort[((i-n0)*ntk)*4+ 3]) {	// if hFlip() == 1
			//alt += PI;
			//phi = PI - ((EMData *)dataref->objectAt(refid-projn0))->phi() + ort[((i-n0)*ntk)*4];
			alt = PI - alt;
			az  = PI + az;
			phi = - ((EMData *)dataref->objectAt(refid-projn0))->phi() + ort[((i-n0)*ntk)*4];
		}
		//else {
		//	phi = ((EMData *)dataref->objectAt(refid-projn0))->phi() + ort[((i-n0)*ntk)*4];
		//}
		float x0 = - ort[((i-n0)*ntk)*4+ 1] * shrink + d->xSize() / 2;
		float y0 = - ort[((i-n0)*ntk)*4+ 2] * shrink + d->ySize() / 2;
				
		if (ver) {
			printf("%d> %s %d ->\t%d\torientation=%6.2f,%6.2f,%6.2f\t=>\t%6.2f,%6.2f,%6.2f\tcenter=%6.2f,%6.2f\t=>\t%6.2f,%6.2f\n",
					num,imgid,i,proj_id,d->alt()*180./PI, d->az()*180./PI,d->phi()*180./PI,alt*180./PI,az*180./PI,phi*180./PI,d->get_center_x(),d->get_center_y(),x0,y0);
			fflush(stdout);
		}
		d->setRAlign(alt,az,phi);
		d->set_center_x(x0);
		d->set_center_y(y0);
		d->writeImage(argv[1],i,1);	// write header only
	}
}
if(scorefile && projbatch == projbatches-1) {
	printf("%d> write to score file %s\n",num,scorefile);
	FILE *scorefp = fopen(scorefile,"w");
	fprintf(scorefp,"#LST\n");
	for (i=n0; i<n1; i++) {
		if (zapped && zapped[i]) continue;	// This particle has been marked as bad
		for (j=0; j<ntk; j++) {
			if (rns[(i-n0)*ntk+j]<0||(rns[(i-n0)*ntk+j]>nr&&(!tree1)&&(!tree2))) {
				printf("%d> Error classifying %d (%d, %g)\n",num,i,rns[(i-n0)*ntk+j],dots[(i-n0)*ntk+j]);
				continue;
			}
			int refid = rns[(i-n0)*ntk+j];	// the reference id in dataref list (can be different from projections index due to tree options)
			int proj_id = rn[refid];	// only take the best score
			//printf("particle %d: refid=%d\tprojid=%d\n", i, refid, proj_id);
			float alt = ((EMData *)dataref->objectAt(refid-projn0))->alt();
			float az  = ((EMData *)dataref->objectAt(refid-projn0))->az();
			float phi = ((EMData *)dataref->objectAt(refid-projn0))->phi() + ort[((i-n0)*ntk)*4];
			if (ort[((i-n0)*ntk)*4+ 3]) {	// if hFlip() == 1
				alt = PI - alt;
				az  = PI + az;
				phi = - ((EMData *)dataref->objectAt(refid-projn0))->phi() + ort[((i-n0)*ntk)*4];
			}
			float x0 = - ort[((i-n0)*ntk+j)*4+ 1] * shrink + d->xSize() / 2;
			float y0 = - ort[((i-n0)*ntk+j)*4+ 2] * shrink + d->ySize() / 2;
			fprintf(scorefp,"%d\t%s\teuler=%g,%g,%g\tcenter=%g,%g\tscore=%g\trefid=%d\tclsinfo=%1.2f,%1.6f,%1.3f,%1.3f,%0.0f\n",
				i,argv[1],alt*180./PI,az*180./PI,phi*180./PI,x0,y0,dots[(i-n0)*ntk+j],proj_id, dots[(i-n0)*ntk+j],ort[((i-n0)*ntk+j)*4],
				ort[((i-n0)*ntk+j)*4+1]*shrink,ort[((i-n0)*ntk+j)*4+2]*shrink,ort[((i-n0)*ntk+j)*4+3]);
		}
	}
	fclose(scorefp);
}

delete d;
for(int ref=0; ref<dataref->count(); ref++) delete (EMData*)dataref->objectAt(ref);
delete dataref;
if(ctfc) {
	for(int ref=0; ref<datarefnc->count(); ref++) delete (EMData*)datarefnc->objectAt(ref);
	delete datarefnc;
}
}	// end  of for(int projbatch=0; projbatch<projbatches; projbatch++)

if (split) {
	printf("%d> write results\n",num);
	for (i=n0; i<n1; i++) {
		if (zapped && zapped[i]) continue;	// This particle has been marked as bad
		for (j=0; j<ntk; j++) {
			if (rns[(i-n0)*ntk+j]<0||(rns[(i-n0)*ntk+j]>nr&&(!tree1)&&(!tree2))) {
				printf("%d> Error classifying %d (%d, %g)\n",num,i,rns[(i-n0)*ntk+j],dots[(i-n0)*ntk+j]);
				continue;
			}
			sprintf(s,"%scls%04d.lst",tree1?"t":"",rn[rns[(i-n0)*ntk+j]]);
			sprintf(t,"%1.2f,  %1.6f,%1.3f,%1.3f,%0.0f",dots[(i-n0)*ntk+j],ort[((i-n0)*ntk+j)*4],
				ort[((i-n0)*ntk+j)*4+1]*shrink,ort[((i-n0)*ntk+j)*4+2]*shrink,ort[((i-n0)*ntk+j)*4+3]);
			d->writeLST(s,-1,argv[1],i,t);
//printf("%d> %d -> %s  WL\n",num,i,s);

			if (logit) {
				// order: program tag, particle number, ntk rank, quality, class, iteration tag
				char s[120];
				sprintf(s,"C\t%d\t%d\t%1.5g,%1.6f,%1.3f,%1.3f,%0.0f\t%d\t%s\n",i,j,dots[(i-n0)*ntk+j],ort[((i-n0)*ntk+j)*4],
					ort[((i-n0)*ntk+j)*4+1]*shrink,ort[((i-n0)*ntk+j)*4+2]*shrink,ort[((i-n0)*ntk+j)*4+3],rn[rns[(i-n0)*ntk+j]],logit);
				FSsend(FS_TEXT,strlen(s)+1,"particle.log",s);
			}
		}
	}
	
	if (logit) { // save log info so that next run can ignore the already done parts
		// order: iteration tag, current fraction, total fractions
		char s[120];
		sprintf(s,"Cycle %s\tjob %d/%d is done\n", logit, num, denom);
		FSsend(FS_TEXT,strlen(s)+1,"classesbymra.log",s);
	}
}

if (touch) {
	FILE *fp = fopen(touch,"wb");
	fclose(fp);
}

if (dfilt) {
	free( c1 );
	free( c2 );
}
free( rn );
free( rns );
free( sim );
free( bests );
free( dots );
free( ort );
free( snr );
delete msk;
delete msk2;

if (!quiet) LOG(Ref,NULL,LOG_END,NULL);
	
}



