// make3d.C 1/98 Steve Ludtke
// This program takes a set of projections (classes) with euler angles and constructs
// a 3d model in fourier space using a gaussian interpolation algorithm.

//N make3d
//F This program builds 3D models from a set of class averages using Fourier techniques
//T LR
//1
//P <input file>	File containing class averages
//P [out=<outfile>]	Output 3D MRC file
//P [hard=<maxpr>]	This specifies how well the class averages must match the model to be included, 25 is typical
//P [noweight]	Normally the class averages are weighted by the number of raw particles used, this disables that
//P [lowmem]	Read in images only when needed instead of preloading all in memory
//P [sym=<c4,icos,etc>]	The symmetry to apply to the model 
//P [pad=<padsize>]	To reduce Fourier artifacts, the model is typically padded by ~25%
//P [mode=<1-6>]	Specifies the interpolation size, 2 is the default
//P [mask=<rad>]	real-space mask radius
//P [apix=<num>]	set the sampling (angstrom/pixel)
//P [norm]	This will perform normalization of the volumes between iterations to eliminate problems in cases with poor coverage in Fourier space
//P [keep=<sigmamult>]	An alternative to 'hard'
//P [resmap=<name>]	Generates a 'resolution map' in another 3D MRC file
//P [snrfile=<file>]	Use a SNR file (as generated by classalignall) for proper 3D Wiener filtration
//P [usecenter]	Will use the center parameter in the image header to center the particle before reconstruction
//P [flipphase]	Will flip the phase to correct CTF
//P [wiener3d=<sf file>]	Will perform Wiener filter in 3D assuming each image has proper CTF parameters embedded
//P [iter=<n>]	Number of iterations, default=4
//P	[subset=<i/n>]	Which subset to use. default = 0/1
//P [goodbad]	Saves the used and unused class averages in 2 files
//P [log]	Averages the log of the projections
//P [savenorm]	Saves the normalization map to norm.mrc
//P [inorm]	This will use a special weighting scheme to compensate for poor SNR sampling on the unit sphere
//U make3d classes.1.img out=threed.1.mrc hard=25 sym=c4 pad=128 mask=42

//D This program is used to build a 3d model from a set of projections/class
//E averages. The individual projections must already have assigned euler
//E angles in their headers. The 3d model building process is iterative. The
//E iterative loop handles 2 things. First, the relative normalization of
//E the slices is not initially known. Second, some class averages will
//E typically be worse than others. Just like when generating class averages
//E from individual images, the images which are the least consistent with
//E the overall model will be excluded from the reconstruction. Similar
//E to classalignall, a sigma multiplier may be specified for discarding
//E individual class averages, or alternatively, a hard phase error
//E threshold may be set. Any image with a higher phase error with the
//E 3d model than the threshold will be discarded. Only one of these 2
//E options, 'keep=<sigmamult>' or 'hard=<max phase error>' should be
//E specified. The phase error should not be interpreted as a residual.
//E Typically this value should be between 20 and 30.

//D 'sym=<symmetry>' specifies the symmetry of the 3d model. If nothing is
//E specified, the model is assumed to have no symmetry. Currently, only
//E Cn, Dn and icos symmetries are supported.

//D Reconstructions are performed in Fourier space. This causes certain
//E artifacts to appear in the 3d model. These artifacts can be reduced
//E by padding the images in real space. 50% padding produces excellent
//E results in most cases, ie - an 80x80 image would be padded to 120x120.
//E This would be specified with 'pad=120'. 25% padding provides adequate
//E results in most cases. The other artifact is caused by interpolation
//E in Fourier space. 'make3d' supports 6 levels of interpolation. Type
//E 1 is 'nearest neighbor' interpolation and is very fast, but produces
//E significant artifacts. The default in most cases is type 2, which 
//E generally produces the best
//E all-around results. If you have a very small number of slices, you
//E might try mode=5. This effect can be observed qualitatively using the 
//E <a href=v4.html>v4</a> program.
//E This program allows you to interactively select the interpolation
//E mode used for generating forward projections of a 3d model. The artifacts
//E in v4 are more severe, but qualitatively similar to those produced
//E by Fourier reconstruction.

//D Finally, since Fourier reconstruction DOES produce some undesirable
//E artifacts, generally this is just used as a preliminary model. Once
//E this model has been generated, typically this model will be improved
//E using a real-space iterative technique. This is implimented in
//E <a href=make3diter.html>make3diter</a>.

#include <stdio.h>
#include "EMData.h"
#include "List.h"
#include "util.h"
#include "Euler.h"
#include <math.h>
#include <ctype.h>

#ifdef MPI
	#include <mpi.h>
#endif

main(int argc, char *argv[]) {
int mode=2,mask=-1,goodbad=0,rsym=1,pad=0,nowt=0,dolog=0,inorm=0;
float sigm=3.0,hard=0,donorm=0;
List *data,*dataf;
EMData *d,*d2,*f,*out,*out2;
int i,j,k,g,imc=0,timc=0,n,ny,ny2,savenorm=0,niter=4,subset=0,nsets=1;
float norm,pr,thr;
float prm,prs,prn;
char *file="threed.mrc",*sym="c1",*resmap=NULL,*snrfile=NULL,*sffile=NULL;
Euler e,*a;
float **SNR=NULL;
float apix=-1.0, apix_img = -1.0; // negative means that the value is not specified by user
bool lowmem = false, usecenter = false, flipphase = false, wiener3d = false;
XYData* SF=NULL;
int myrank=0, ncpus=1;

#ifdef MPI
	char cwd[256];
	getcwd(cwd,255);
	MPI_Init(&argc,&argv);
	MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
	MPI_Comm_size(MPI_COMM_WORLD,&ncpus);
	if(ncpus==1) chdir(cwd);
#endif

appinit(argc,argv);

if (argc<2) { appusage(argv[0]); exit(1); }

n=fileCount(argv[1]);
if (n<3) { printf("Input file too small ?!?! only %d images in %s\n",n,argv[1]); exit(1); }

for (i=2; i<argc; i++) {	
	if (strncmp(argv[i],"out=",4)==0) file=&argv[i][4];
	else if (strncmp(argv[i],"mode=",5)==0) mode=atoi(&argv[i][5]);
	else if (strncmp(argv[i],"hard=",5)==0) hard=atof(&argv[i][5])*PI/180.0;
	else if (strncmp(argv[i],"mask=",5)==0) mask=atoi(&argv[i][5]);
	else if (strncmp(argv[i],"apix=",5)==0) apix=atof(&argv[i][5]);
	else if (strncmp(argv[i],"resmap=",7)==0) resmap=&argv[i][7];
	else if (strncmp(argv[i],"keep=",5)==0) sigm=atof(&argv[i][5]);
	else if (strncmp(argv[i],"goodbad",7)==0) goodbad=1;
	else if (strncmp(argv[i],"inorm",5)==0) inorm=1;
	else if (strncmp(argv[i],"savenorm",8)==0) savenorm=1;
	else if (strncmp(argv[i],"log",3)==0) dolog=1;
	else if (strncmp(argv[i],"norm",4)==0) donorm=1;
	else if (strncmp(argv[i],"snrfile=",8)==0) snrfile=argv[i]+8;
	else if (strncmp(argv[i],"noweight",8)==0) nowt=1;
	else if (strncmp(argv[i],"lowmem",6)==0) lowmem=1;
	else if (strncmp(argv[i],"usecenter",9)==0) usecenter=1;
	else if (strncmp(argv[i],"flipphase",9)==0) flipphase=1;
	else if (strncmp(argv[i],"wiener3d=",9)==0) { 
		wiener3d=1; 
		sffile=argv[i]+9; 
		SF = new XYData();
		if (SF->readFile(sffile)) {
			fprintf(stderr,"ERROR: Couldn't read structure factor file %s!\n", sffile);
			exit(-1);
		}
		float *d0=SF->getData();
		for (j=1; j<SF->n()*2; j+=2) d0[j]=(d0[j]<=0)?-4.0:log10(d0[j]);
		SF->update();
	}
	else if (strncmp(argv[i],"iter=",5)==0) {
		niter=atoi(&argv[i][5]);
		if(niter<=0) niter=1;
	}
	else if (strncmp(argv[i],"subset=",7)==0) {
		sscanf(&argv[i][7],"%d/%d",&subset,&nsets);
		if (nsets<=0) nsets=1;
		if (subset<0 || subset>=nsets) subset=0;
	}
	else if (strncmp(argv[i],"sym=",4)==0) {
		sym=&argv[i][4];
		Euler e;
		e.setSym(sym);
		if (e.getMaxSymEl()==0) { printf("Unsupported symmetry\n"); exit(1); }
	}
	else if (strncmp(argv[i],"pad=",4)==0) pad=atoi(&argv[i][4]);
	else if (strncmp(argv[i],"usefs",5)==0) continue;
	else { printf("Unknown argument '%s'\n",argv[i]); exit(1); }
}

#ifdef MPI
	subset+=myrank*nsets;
	nsets*=ncpus;
#endif

if (goodbad) {
	unlink("3dgood.lst");
	unlink("3dbad.lst");
}

int Ref;
if (myrank==0){
	Ref=LOGbegin(argc,argv);
	LOG(Ref,argv[1],LOG_INFILE,NULL);
}

dataf=new List;

// Read slices
printf("Read data\n");
data=readImages(argv[1],0,-1,lowmem);	// header only if lowmem==1
if (data->count()<3) {
	printf("Need at least 3 images!\n");
	exit(1);
}

d=(EMData *)data->objectAt(0);
ny=d->ySize();
ny2=ny;
if (pad && pad > ny) ny2=pad;
else if (pad && pad < ny) {
	fprintf(stderr,"ERROR in make3d: pad=%d is smaller than the original image size %d", pad, ny);
	exit(2);
}

// trying to get the sampling right
if(d->hasCTF()) apix_img = (d->getCTF())[10];
else apix_img = d->Pixel();
if(apix<=0 && apix_img>0) apix = apix_img;  

// prepare slices
printf("Transform slices\n");
for (i=subset; i<data->count(); i+=nsets) {
	d=(EMData *)data->objectAt(i);
	if (d->NImg()<0||d->NImg()>1000000) { delete d; continue; }
	if (d->NImg()==0) d->setNImg(1);
	timc+=d->NImg();
	if(lowmem) {
		d->unum4=i;
		dataf->addObject(d);
	}
	else {
		d->normalize();
		d->applyMask(mask,6);
		d->getEuler()->setSym(sym);

		d2=d->clip((ny-ny2)/2,(ny-ny2)/2,ny2,ny2);
		f=d2->doFFT();
		f->toCorner();
		d2->gimmeFFT();
		f->unum4=i;
		dataf->addObject(f);
		f->setRAlign(d->alt(),d->az(),d->phi());
		f->setTAlign( d->get_center_x()-ny/2, d->get_center_y()-ny/2, 0);
		f->set_center_x( d->get_center_x()-ny/2+ny2/2);
		f->set_center_y( d->get_center_y()-ny/2+ny2/2);
		f->setNImg(d->NImg());
		if(d->hasCTF()) f->setCTF(d->getCTF());

		delete d2;
		delete d;
	}
}
delete data;

if (snrfile) {
	SNR = (float **)malloc(dataf->count()*sizeof(float *));
	for (i=subset; i<dataf->count(); i+=nsets) {
		d=(EMData *)dataf->objectAt(i);
		EMData *tmp = new EMData;
		if (tmp->readImage(snrfile,d->unum4)) error(ERR_EXIT,"Error reading SNR data");
		SNR[i]=tmp->getDataRO();
		// now we just let the EMData object pointer drop, but keep the data pointer.
		// this is very bad style, but will work ok ...
	}
}

out=new EMData;
out->setup4IS(ny2);
out2=NULL;

if(lowmem) d = new EMData;

thr=PI/2.0;
for (j=0; j<niter; j++) {
//	if (j==3) mode=6;
	if (hard>0) {
		if (niter>1) thr=hard*(1.0+float(niter-1-j)/float(niter-1));	// gradually enforce the hard threshold
		else thr = hard;
	}

	imc=0;
	prm=prs=prn=0;
	for (i=0; i<dataf->count(); i++) {
		int real_index = ((EMData *)dataf->objectAt(i))->unum4;
		g=1;
		if(lowmem) {
			d->readImage(argv[1],((EMData *)dataf->objectAt(i))->unum4);
			if (d->NImg()==0) d->setNImg(1);
			d->normalize();
			if (usecenter) {
				float dx = d->get_center_x() - d->ySize()/2;
				float dy = d->get_center_y() - d->ySize()/2;
				d->applyMask(mask,6,dx,dy,0);
			}
			else {
				d->applyMask(mask,6);
			}
				
			d->getEuler()->setSym(sym);

			d2=d->clip((ny-ny2)/2,(ny-ny2)/2,ny2,ny2);
			f=d2->doFFT();
			f->toCorner();
			d2->gimmeFFT();
			f->unum4=i;
			f->setRAlign(d->alt(),d->az(),d->phi());
			f->setTAlign( d->get_center_x()-ny/2, d->get_center_y()-ny/2, 0);
			f->set_center_x( d->get_center_x()-ny/2+ny2/2);
			f->set_center_y( d->get_center_y()-ny/2+ny2/2);
			f->setNImg(d->NImg());
			if(d->hasCTF()) f->setCTF(d->getCTF());
			delete d;
			delete d2;
			d=f;
		}
		else {
			d=(EMData *)dataf->objectAt(i);
		}
		if (d->NImg()<=0) continue;
		if (nowt) d->setNImg(1);

		// phase shift to compensate center shift
		if(usecenter && (lowmem || (!lowmem && j==0)) ) {
			float dx = d->get_center_x() - d->ySize()/2;
			float dy = d->get_center_y() - d->ySize()/2;
			for ( int x = 0; x < ny2/2; x++ ) {
				for ( int y = 0; y < ny2; y++ ) {
					float r0 = d->valueAt(2 * x + 0, y, 0 );
					float i0 = d->valueAt(2 * x + 1, y, 0 );
	
					float PhaseShift = 2.0F * 3.1415926535897932384626433832795F *
									(
										(
											( x         ) * dx +
											( y - ny2/2 ) * dy 
										)
										/ float ( ny2 ) 
									);
					float psr = cos(PhaseShift);
					float psi = sin(PhaseShift);

					// Perform the phase shifting:
					d->setValueAt ( 2 * x + 0, y, 0, r0*psr - i0*psi );
					d->setValueAt ( 2 * x + 1, y, 0, r0*psi + i0*psr );
				}
			}
		}
		// flip phase 
		if(flipphase && (lowmem || (!lowmem && j==0)) ) {
			if (d->hasCTF()) {
				EMData *filter = new EMData();
				filter->setCTF(d->getCTF());
				filter->setSize(d->xSize(),d->ySize(),1);
				filter->setComplex(1);
				filter->setRI(1);
				filter->ctfMap(flipphase*2);
				d->mult(filter);
				delete filter;
			}
			else {
				fprintf(stderr,"WARNING: image %d does not have CTF parameter in the header\n",i);
			}
		}

		if (j==niter-1 && dolog) d->realFilter(15);
		
		if (j!=0 && (j!=niter-1||!dolog)) {
			d2=d;
			norm=out2->normSlice(d2,d2->alt(),d2->az(),d2->phi(),&pr);
			if (norm!=0) { d2->multConst(1.0/norm); }
			
			prm+=pr;
			prs+=pr*pr;
			prn+=1.0;
			printf("%d/%d(%d)  %d\t%5.1f  %5.1f  %5.1f\t%f\t%f ",i,dataf->count(),
				real_index,d2->NImg(),d2->alt()*180.0/PI,d2->az()*180.0/PI,d2->phi()*180.0/PI,
				norm,pr*180.0/PI); 
			if (pr>thr||norm<.01||norm>100.0) { g=0; printf(" X\n"); }
			else printf("\n");
		}
		else printf("%d/%d(%d)  %d\t%5.1f  %5.1f  %5.1f\t\t%6.2g %6.2g\n",i,dataf->count(),
				real_index,d->NImg(),d->alt()*180.0/PI,d->az()*180.0/PI,d->phi()*180.0/PI,
				d->Max(),d->Min()); 

		EMData *filter;
		if (d->NImg()>0 && g) {
			if(wiener3d && d->hasCTF()) {
				filter = new EMData();
				filter->setCTF(d->getCTF());
				filter->setSize(d->xSize(),d->ySize(),1);
				filter->setComplex(1);
				filter->setRI(1);
				filter->ctfMap(16, SF);	//absolute SNR
			}
			imc+=d->NImg();
/*			for (k=0; k<rsym; k++) {
				out->insertSlice(d,d->alt(),d->az()+k*2.0*PI/(float)rsym,
					d->phi(),mode,d->NImg()); 
			}*/
			a=d->getEuler();
			for (e=a->FirstSym(); e.valid(); e=a->NextSym()) {
//printf("\t%f\t%f\t%f\n",e.alt()*180./PI,e.az()*180./PI,e.phi()*180./PI);
				if (SNR) out->insertSliceWF(d,e.alt(),e.az(),e.phi(),mode,SNR[i],ny2/(float)ny, 1, 1);
				else if(wiener3d && d->hasCTF()) {
					out->insertSliceWF(d,e.alt(),e.az(),e.phi(),mode,filter->getDataRO(), 1.0, 0, 2);
					filter->doneData();
				}
				else out->insertSlice(d,e.alt(),e.az(),e.phi(),mode,d->NImg()); 			
			}
			if(wiener3d && d->hasCTF()) delete filter;
		}
		if (j==niter-1 && goodbad && d->NImg()>0) {
			if (g) d->writeLST("3dgood.lst",-1,argv[1], ((EMData *)dataf->objectAt(i))->unum4);
			else d->writeLST("3dbad.lst",-1,argv[1], ((EMData *)dataf->objectAt(i))->unum4);
		}
	}
	
	if (prn) {
		prm/=prn;
		prs=sqrt(prs/prn-prm*prm);
		printf("Phase error Mean=%f   Sigma=%f   used %d/%d images\n",prm*180.0/PI,prs*180.0/PI,imc,timc);
		thr=prm+prs*sigm;
	#ifdef MPI
		if(ncpus>1)    {
			int total_num_imgs_used=0;
			int total_num_imgs=0;
			MPI_Reduce(&imc, &total_num_imgs_used, 1,  MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD);
			MPI_Reduce(&timc, &total_num_imgs, 1,  MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD);
			if(myrank==0) printf("Total number of images used: %d/%d\n", total_num_imgs_used,total_num_imgs);
		}
	#endif
	}

	if (out2) {
		if(out2->Parent()) delete out2->Parent();
		delete out2;
		out2=0;
	}

#ifdef MPI
	if(ncpus>1)    {
		int nelements = ny2*(ny2+2), sections = ny2;
		float *fft_send_dp = out->getData();
		float *fft_recv_dp = 0;
		if (myrank==0) {
			fft_recv_dp = calloc(nelements, sizeof(float));
			if( ! fft_recv_dp ) {
				printf("ERROR: failed to allocate memory\n");
				exit(1);
			}
		}
		for(int section=0; section < sections; section++) { // by section instead of whole map, to save memory
			if (myrank==0) memset(fft_recv_dp, 0, sizeof(float)*nelements);
			MPI_Reduce(&fft_send_dp[section*nelements], fft_recv_dp, nelements,  MPI_FLOAT, MPI_SUM, 0, MPI_COMM_WORLD);
			if (myrank==0) memcpy(&fft_send_dp[section*nelements],fft_recv_dp,sizeof(float)*nelements);
		}
		
		float *wt_send_dp = out->Parent()->getData();
		float *wt_recv_dp = 0;
		if (myrank==0) wt_recv_dp = fft_recv_dp;
		for(int section=0; section < sections; section++) { 
			if (myrank==0) memset(wt_recv_dp, 0, sizeof(float)*nelements);
			MPI_Reduce(&wt_send_dp[section*nelements], wt_recv_dp, nelements, MPI_FLOAT, MPI_SUM, 0, MPI_COMM_WORLD);
			if (myrank==0) memcpy(&wt_send_dp[section*nelements],wt_recv_dp,sizeof(float)*nelements);
		}
		if (myrank==0) free(fft_recv_dp);

		if(j<niter-1) {
			MPI_Bcast(fft_send_dp, nelements, MPI_FLOAT, 0, MPI_COMM_WORLD);
			MPI_Bcast(wt_send_dp, nelements, MPI_FLOAT, 0, MPI_COMM_WORLD);
		}
		out->doneData();
		out->Parent()->doneData();
	}
#endif

	if (SNR || wiener3d) out->doneSliceWF();
	else out->doneSlice(0);
	out2=out;
	if (donorm) {
		out2->normalize();
	}
	if(j<niter-1) {
		out=new EMData;
		out->setup4IS(ny2);
	}
}

for(i=0; i< dataf->count(); i++) delete (EMData*)(dataf->objectAt(i));
delete dataf;

if (goodbad && myrank==0) {
	LOG(Ref,"3dgood.hed",LOG_OUTFILE,NULL);
	LOG(Ref,"3dbad.hed",LOG_OUTFILE,NULL);
}

if(myrank==0) {
	//if (SNR || wiener3d) out2->doneSliceWF(dolog);
	//else out2->doneSlice(dolog);
	if (savenorm) {
		EMData nm;
		nm.setSize(out2->xSize()/2,out2->ySize(),out2->ySize());
		float *d=nm.getData();
		for (i=0; i<out2->xSize()*out2->ySize()*out2->zSize(); i+=2) {
			d[i/2]=out2->Parent()->valueAt(i,0,0);
		}
		nm.doneData();
		nm.writeMRC("norm.mrc");
		d=nm.getData();
		for (i=0; i<out2->xSize()*out2->ySize()*out2->zSize(); i+=2) {
			d[i/2]=hypot(out2->valueAt(i,0,0),out2->valueAt(i+1,0,0));
		}
		nm.doneData();
		nm.writeMRC("map.mrc");
	
	}
	if (dolog) out2->realFilter(16);
	out2->toCorner();
	
	if (resmap) {
		float *d1,*d2;
		
		f=out2->Parent();
		out->setSize(ny2,ny2,ny2);
		out->zero();
		d1=f->getDataRO();
		d2=out->getData();
		
		for (i=0; i<ny2; i++) {
			for (j=0; j<ny2; j++) {
				for (k=ny2/2; k<ny2; k++) {
					d2[k+j*ny2+i*ny2*ny2]=sqrt(d1[(k-ny2/2)*2+j*(ny2+2)+i*(ny2+2)*ny2]);
					if (k!=ny2/2 && i!=0 && j!=0) {
						d2[(ny2-k)+(ny2-j)*ny2+(ny2-i)*ny2*ny2]=sqrt(d1[(k-ny2/2)*2+j*(ny2+2)+i*(ny2+2)*ny2]);
					}
				}
			}
		}
		f->doneData();
		out->doneData();
	//	out=out->clip((ny2-ny)/2,(ny2-ny)/2,(ny2-ny)/2,ny,ny,ny);
		out->writeMRC(resmap);
		
		LOG(Ref,resmap,LOG_INFILE,NULL);
	
	}
	
	// If we applied a 3D wiener filter, write the filter to a file
	if (SNR|| wiener3d) {
		f=out2->Parent();
		float *d1=f->getData();
		for (i=0; i<ny2*ny2*(ny2+2); i+=2) d1[i]=d1[i+1]=1./(1.+(1./d1[i]));
		f->doneData();
		f->writeMRC("filter3d.mrc");
		LOG(Ref,"filter3d.mrc",LOG_OUTFILE,NULL);
	}
}

delete out2->Parent();

if(myrank==0) {
	out=out2->doIFT();
	
	out2->gimmeFFT();
	delete out2;
	
	//out->applyMask(-1,100+mode);
	if (pad) {
		out2=out->clip((ny2-ny)/2,(ny2-ny)/2,(ny2-ny)/2,ny,ny,ny);
		delete out;
		out=out2;
	}
	out->applyMask(mask,6);
	
	// now we apply d symmetry 2-fold at the end, note 2f must be on x-axis
	/*if (tolower(sym[0])=='d') {
		EMData *tmp;
		tmp=out->copy(0,0);
		tmp->setRAlign(PI,0,0);
		tmp->rotateAndTranslate();
		out->add(tmp);
		delete tmp;
	}*/
	
	// record the sampling and origin in the map header so that later segmentation and visualization
	if(apix>0) {
		out->setPixel(apix);
		out->setXYZOrigin( -out->xSize()/2 * apix,  -out->ySize()/2 * apix, -out->zSize()/2 * apix);
	}
	
	out->writeMRC(file);
	
	LOG(Ref,file,LOG_OUTFILE,NULL);
	
	delete out;
	
	LOG(Ref,NULL,LOG_END,NULL);
}
#ifdef MPI
	MPI_Finalize();
#endif
}
